#!/bin/sh
#Script to notify the website about a release

function getVersionToRelease {
    CURRENT_VERSION=`mvn org.apache.maven.plugins:maven-help-plugin:2.1.1:evaluate -Dexpression=project.version | grep -v "\[INFO\]"`
    echo ${CURRENT_VERSION%%-SNAPSHOT}
}

function getVersion {
    mvn org.apache.maven.plugins:maven-help-plugin:2.1.1:evaluate -Dexpression=project.version | grep -v "\[INFO\]"
}
function getBranch {
    git symbolic-ref --short HEAD
}

case $1 in
request)
    echo "Start release with request option on"
    VERSION=$(getVersionToRelease)
    echo ${VERSION}
    RELEASE_BRANCH="release-$VERSION"
    git checkout -b ${RELEASE_BRANCH} || exit 1

    echo "-- Prepare the code for release ${VERSION} in branch ${RELEASE_BRANCH} --"
    echo $VERSION > .version
    git add .version
    sh ./bump_release.sh code $VERSION || exit 1

    git commit -m "Prepare the code for release ${VERSION}" -a
    git push origin ${RELEASE_BRANCH} || exit 1
    git checkout develop
    echo "Branch $RELEASE_BRANCH is ready for the releasing process"
    ;;
perform)
    echo "Start release with perform option on"
    if [ ! -f .version ]; then
        echo "Missing .version file"
        exit 1
    fi
    VERSION=$(cat .version)

    echo "-- Prepare the release --"
    if ! mvn -B release:prepare; then
        mvn release:rollback
        exit 1
    fi

    echo "-- Perform the release --"
    if ! mvn release:perform; then
        mvn release:rollback
        exit 1
    fi
    rm .version #To prevent for an infinite loop

    DEV_HEAD=$(git rev-parse HEAD)
    RELEASE_BRANCH="release-${VERSION}"
    # The current tree looks like:
    # * HEAD -> next dev version, it is a detached head due to jenkins git plugin
    # * (tag ...) -> the tag made by maven release:prepare on the relased version == ${DEV_HEAD}~1
    # * ($RELEASE_BRANCH) -> the pointer on the released branch that was set by the client with ./release.sh request

    # merge the version changes back into develop so that folks are working against the new release
    echo "-- Integrate the next version into the develop branch --"
    git checkout develop
    NEW_VERSION=$(getVersion)
    echo "-- Integrate the next version ($DEV_HEAD) into the develop branch --"
    git commit -m "integrate the previous development version ${VERSION} generated by maven" || exit 1
    git merge --no-ff -m "integrate the development version ${NEW_VERSION} generated by maven" ${DEV_HEAD} || exit 1
    ./bump_release.sh code ${NEW_VERSION}
    git commit -m "code prepared for development version ${NEW_VERSION}" -a


    echo "-- Integrate release ${VERSION} into the master branch --"
    #git checkout ${RELEASE_BRANCH}
    #git merge -s ours master -m "merge master into the release"
    git checkout master
    git merge --no-ff ${DEV_HEAD}~1 -m "integrate release ${VERSION} to master"

    echo "-- Generate the javadoc for release ${VERSION} --"
    mvn javadoc:aggregate > /dev/null

    echo "-- Push the changes and the tags --"
    git branch -D ${RELEASE_BRANCH}
    git push --all && git push --tags
    git push origin --delete ${RELEASE_BRANCH}

    ;;
zip)
    echo "Start release with zip option on"
    VERSION=$(cat .version)
    mkdir choco-${VERSION}
    git checkout choco-${VERSION}
    mvn clean install -DskipTests || exit 1
    mv ./choco-solver/target/choco-solver-${VERSION}-jar-with-dependencies.jar ./choco-${VERSION}
    mv ./choco-solver/target/choco-solver-${VERSION}-sources.jar ./choco-${VERSION}
    mv ./choco-samples/target/choco-samples-${VERSION}-sources.jar ./choco-${VERSION}
    mv ./choco-parser/target/choco-parser-${VERSION}-jar-with-dependencies.jar ./choco-${VERSION}
    mv ./user_guide.pdf ./choco-${VERSION}/user_guide-${VERSION}.pdf

    mvn javadoc:aggregate  || exit 1
    cd target/site/
    zip -r apidocs-${VERSION}.zip ./apidocs/*
    mv apidocs-${VERSION}.zip ../../choco-${VERSION}
    cd ../../

    zip choco-${VERSION}.zip ./choco-${VERSION}/*

    git checkout develop
    rmdir choco-${VERSION}
    ;;
    *)
        echo "Unsupported operation '$1'"
        exit 1
esac
